# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import glob
import os
import string
import unicodedata
import json
import math
import operator
from functools import reduce

import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sn

datafiles = "./data/names/*.txt"  # les fichiers pour construire vos modèles
test_filename = './data/test-names-t2.txt'  # le fichier contenant les données de test pour évaluer vos modèles

names_by_origin = {}  # un dictionnaire qui contient une liste de noms pour chaque langue d'origine
all_origins = []  # la liste des 18 langues d'origines de noms

BOS = "~"  # character used to pad the beginning of a name
EOS = "!"  # character used to pad the end of a name
v = 26
kgram_models = {}


def find_files(path):
    """Retourne le nom des fichiers contenus dans un répertoire.
       glob fait le matching du nom de fichier avec un pattern - par ex. *.txt"""
    return glob.glob(path)


def get_origin_from_filename(filename):
    """Passe-passe qui retourne la langue d'origine d'un nom de fichier.
       Par ex. cette fonction retourne Arabic pour "./data/names/Arabic.txt". """
    return os.path.splitext(os.path.basename(filename))[0]


def unicode_to_ascii(s):
    """Convertion des caractères spéciaux en ascii. Par exemple, Hélène devient Helene.
       Tiré d'un exemple de Pytorch. """
    all_letters = string.ascii_letters + " .,;'"
    return ''.join(
        c for c in unicodedata.normalize('NFD', s)
        if unicodedata.category(c) != 'Mn'
        and c in all_letters
    )


def read_names(filename):
    """Retourne une liste de tous les noms contenus dans un fichier."""
    with open(filename, encoding='utf-8') as file_handle:
        names = file_handle.read().strip().split('\n')
    return [unicode_to_ascii(name) for name in names]


def load_names():
    """Lecture des noms et langues d'origine d'un fichier. Par la suite,
       sauvegarde des noms pour chaque origine dans le dictionnaire names_by_origin."""
    for filename in find_files(datafiles):
        origin = get_origin_from_filename(filename)
        all_origins.append(origin)
        names = read_names(filename)
        names_by_origin[origin] = names


# Cette fonction normalise les mots en les convertissant en ASCII, mettant en minuscules et supprimant les espaces
def normalize_word(word):
    word = unicode_to_ascii(word)
    word = word.lower()
    word = word.replace(" ", "")
    return word


# On génère les unigrammes à l'aide d'une liste
def make_unigram(word):
    return list(word)


# On génère les k-grammes en ajoutant du padding et
# en traversant la chaîne de caractères de gauche à droite
# avec un pas de 1 et une largeur de k
def make_kgram(unigram_word, k):
    # Ajouter padding
    unigram_with_padding = [BOS for i in range(1, k)] + unigram_word + [EOS for i in range(1, k)]
    kgrams = [''.join(unigram_with_padding[i:i + k]) for i in range(0, len(unigram_with_padding) - k + 1)]
    return kgrams


# Cette fonction crée un modèle de langue pour une origine en prenant en entrée des k-grammes
def make_count_dict(kgrams_origin):
    kgram_count_dict = {}
    for kgram_name in kgrams_origin:
        for kgram in kgram_name:
            kgram_count_dict[kgram] = kgram_count_dict.get(kgram, 0) + 1
    return kgram_count_dict


def train_models():
    load_names()
    # Vous ajoutez à partir d'ici tout le code dont vous avez besoin
    # pour construire vos modèles N-grammes pour chacune des langues d'origines.
    #
    # Vous pouvez ajouter au fichier toutes les fonctions que vous jugerez nécessaire.
    # Merci de ne pas modifier les fonctions présentes dans ce fichier.

    # Construire un tableau de fréquence par langue. Chaque tableau est une entrée dans un dictionnaire
    for origin in names_by_origin.keys():
        unigrams = []
        for word in names_by_origin[origin]:
            word = normalize_word(word)
            unigrams.append(make_unigram(word))
        for k in range(1, 5):
            kgrams_origin = []
            for unigram_word in unigrams:
                kgrams_origin.append(make_kgram(unigram_word, k))
            kgram_models[(origin, k)] = make_count_dict(kgrams_origin)


def most_probable_origin(name, n=3, use_logprob=True):
    # Retourne la langue d'origine la plus probable du nom.
    # n désigne la longueur des N-grammes. Par ex n=3 --> trigramme

    # Calculer les log-prob et les perplexités
    log_prob_origin = {}
    perplexity_origin = {}
    for origin in names_by_origin.keys():
        log_prob_origin[origin] = logprob(name, origin, n)
        perplexity_origin[origin] = perplexity(name, origin, n)

    # Trouver l'optimum de chacun
    most_probable_origin_prob = max(log_prob_origin.items(), key=lambda k: k[1])[0]
    most_probable_origin_perp = min(perplexity_origin.items(), key=lambda k: k[1])[0]

    if use_logprob:
        return most_probable_origin_prob
    else:
        return most_probable_origin_perp


def logprob(name, origin, n=3):
    # Retourne la valeur du logprob d'un nom étant donné une origine
    # Utilisez une fonction logarithme en base 2.

    unigram_name = make_unigram(name)
    kgram_name = make_kgram(unigram_name, n)
    k_counts = kgram_models[(origin, n)]
    logprob_name = 0
    if n == 1:
        prob_n = sum(k_counts.values())
        for kgram in kgram_name:
            prob_c = k_counts.get(kgram, 0)
            prob_l = math.log((prob_c + 1) / (prob_n + v))
            logprob_name = logprob_name + prob_l
    else:
        k1_counts = kgram_models[(origin, n - 1)]
        for kgram in kgram_name:
            prob_n = k1_counts.get(kgram[:-1], 0)
            prob_c = k_counts.get(kgram, 0)
            prob_l = math.log((prob_c + 1) / (prob_n + v))
            logprob_name = logprob_name + prob_l
    return logprob_name


def prod(iterable):
    return reduce(operator.mul, iterable, 1)


def perplexity(name, origin, n=3):
    # Retourne la valeur de perplexité d'un nom étant donné une origine
    # À compléter...
    unigram_name = make_unigram(name)
    kgram_name = make_kgram(unigram_name, n)
    length_name = len(kgram_name)
    k_counts = kgram_models[(origin, n)]
    probs_name = []
    if n == 1:
        prob_n = sum(k_counts.values())
        for kgram in kgram_name:
            prob_c = k_counts.get(kgram, 0)
            prob_p = (prob_c + 1) / (prob_n + v)
            probs_name.append(prob_p)
    else:
        k1_counts = kgram_models[(origin, n - 1)]
        for kgram in kgram_name:
            prob_n = k1_counts.get(kgram[:-1], 0)
            prob_c = k_counts.get(kgram, 0)
            prob_p = (prob_c + 1) / (prob_n + v)
            probs_name.append(prob_p)

    perp = math.exp(-sum([math.log(x) for x in probs_name]) / length_name)

    return perp


def load_test_names(filename):
    """Retourne un dictionnaire contenant les données à utiliser pour évaluer vos modèles.
       Le dictionnaire contient une liste de noms (valeurs) et leur origine (clé)."""
    with open(filename, 'r') as fp:
        test_data = json.load(fp)
    return test_data


def evaluate_models(filename, n=3, use_logprob=True):
    """Fonction utilitaire pour évaluer vos modèles. Aucune contrainte particulière.
       Je n'utiliserai pas cette fonction pour l'évaluation de votre travail. """
    test_data = load_test_names(filename)
    # À compléter - Fonction pour l'évaluation des modèles N-grammes.
    # ...

    results = {}
    compteur_succes = {}
    for org, name_list in test_data.items():
        compteur_succes[org] = {"succes": 0, "echec": 0}
        results[org] = {}
        for name in name_list:
            mp_origin = most_probable_origin(name, n, use_logprob)
            results[org][mp_origin] = results[org].get(mp_origin, 0) + 1
            if org == mp_origin:
                compteur_succes[org]["succes"] = compteur_succes[org].get("succes", 0) + 1
            else:
                compteur_succes[org]["echec"] = compteur_succes[org].get("echec", 0) + 1
    return results, compteur_succes


def compute_v():
    global v
    vocab = {}
    for origin in all_origins:
        modele_courant = kgram_models[(origin, 1)]
        for k in modele_courant.keys():
            vocab[k] = vocab.get(k, 0) + modele_courant.get(k, 0)
    v = len(vocab.keys()) + 1
    print("la valeur de V est " + str(v))


if __name__ == '__main__':
    # Vous pouvez modifier cette section comme bon vous semble
    load_names()
    print("Les {} langues d'origine sont:".format(len(all_origins)))
    print(all_origins)
    print("Les noms chinois sont:")
    print(names_by_origin["Chinese"])

    train_models()
    compute_v()
    some_name = "Lamontagne"
    some_origin = most_probable_origin(some_name)
    some_logprob = logprob(some_name, some_origin)
    some_perplexity = perplexity(some_name, some_origin)
    print("\nLangue d'origine de {}: ".format(some_name), some_origin)
    print("logprob({}, {}):".format(some_name, some_origin), some_logprob)
    print("perplexity({}, {}):".format(some_name, some_origin), some_perplexity)

    test_names = load_test_names(test_filename)
    # print("\nLes données pour tester vos modèles sont:")
    # for org, name_list in test_names.items():
    #    print("\t", org, name_list)
    for current_n in range(1, 4):
        res_logprob, succes_logprob = evaluate_models(test_filename, current_n, True)
        res_perp, succes_perp = evaluate_models(test_filename, current_n, False)

        print(res_logprob)

        # Résultats pour le rapport

        df_succes_logprob = pd.DataFrame.from_dict(succes_logprob).fillna(0).transpose()
        df_succes_logprob.columns = ['echec_logprob', 'succes_logprob']
        df_succes_perp = pd.DataFrame.from_dict(succes_perp).fillna(0).transpose()
        df_succes_perp.columns = ['echec_perp', 'succes_perp']
        df_succ = df_succes_logprob.join(df_succes_perp)

        f, ax = plt.subplots(figsize=(9, 9))
        hm_echec = sn.heatmap(df_succ, annot=True, fmt="d", linewidths=.5, ax=ax, cmap='Oranges')
        hm_echec.set(
            title="Performance de modèles " +
                  {1: "uni", 2: "bi", 3: "tri"}[current_n] +
                  "grammes \nsur les données de test")
        plt.tight_layout()
        hm_echec.get_figure().savefig("images/2_seaborn_hm_echec_" + str(current_n) + ".png")
